@@ -1,12 +1,11 @@
 /* sqlite-module.c
  * S-Lang bindings for the sqlite3 library
- * This was tested with sqlite3 version 3.3.8
+ * This was tested with sqlite3 version 3.5.9
  * 
- * $Id: sqlite-module.c,v 1.5 2007/10/13 12:39:36 paul Exp paul $
- * 
- * Copyright (c) 2006, 2007 Paul Boekholt.
+ * Copyright (c) 2006, 2007, 2010, 2017 Paul Boekholt.
  * Released under the terms of the GNU GPL (version 2 or later).
  */
+#include "config.h"
 #include <sqlite3.h>
 #include <slang.h>
 #include <string.h>
@@ -14,9 +13,9 @@
 SLANG_MODULE(sqlite);
 
 #define MODULE_MAJOR_VERSION	0
-#define MODULE_MINOR_VERSION	4
-#define MODULE_PATCH_LEVEL	0
-static char *Module_Version_String = "0.4.0";
+#define MODULE_MINOR_VERSION	5
+#define MODULE_PATCH_LEVEL	1
+static char *Module_Version_String = "0.5.1";
 #define MODULE_VERSION_NUMBER	\
    (MODULE_MAJOR_VERSION*10000+MODULE_MINOR_VERSION*100+MODULE_PATCH_LEVEL)
 
@@ -31,6 +30,29 @@ typedef struct
 
 #define DUMMY_SQLITE_TYPE 255
 /*}}}*/
+/*{{{ Statement type */
+
+static int Statement_Type_Id = 0;
+typedef struct
+{
+   sqlite3_stmt * ppStmt;
+   /*
+    * The state of the prepared statement determines which functions are permissible.
+    * state		function	end-state
+    * -----------	-------------	------------------------
+    * SQLITE_OK 	step 		SQLITE_ROW or SQLITE_DONE
+    * 			reset		SQLITE_OK
+    * 			bind_param(s)	SQLITE_OK
+    * SQLITE_ROW 	step		SQLITE_ROW or SQLITE_DONE
+    * 			fetch		SQLITE_ROW
+    * 			reset		SQLITE_OK
+    * SQLITE_DONE	reset		SQLITE_OK
+    */
+   int state;
+} Statement_Type;
+
+#define DUMMY_STATEMENT_TYPE 254
+/*}}}*/
 /*{{{ exceptions */
 static int  Sqlite_Error = 0;
 static int  Sqlite_Internal_Error = 0;
@@ -70,32 +92,32 @@ Sqlite_Exception_Table_Type;
 
 static const Sqlite_Exception_Table_Type Sqlite_Exception_Table [] = 
 {
-     { SQLITE_ERROR, &Sqlite_Error, "Sqlite_Error", "Sqlite Error"},
-     { SQLITE_INTERNAL, &Sqlite_Internal_Error, "Sqlite_Internal_Error", "NOT USED. Internal logic error in SQLite"},
-     { SQLITE_PERM, &Sqlite_Perm_Error, "Sqlite_Perm_Error", "Access permission denied"},
-     { SQLITE_ABORT, &Sqlite_Abort_Error, "Sqlite_Abort_Error", "Callback routine requested an abort"},
-     { SQLITE_BUSY, &Sqlite_Busy_Error, "Sqlite_Busy_Error", "The database file is locked"},
-     { SQLITE_LOCKED, &Sqlite_Locked_Error, "Sqlite_Locked_Error", "A table in the database is locked"},
-     { SQLITE_NOMEM, &Sqlite_Nomem_Error, "Sqlite_Nomem_Error", "A malloc() failed"},
-     { SQLITE_READONLY, &Sqlite_Readonly_Error, "Sqlite_Readonly_Error", "Attempt to write a readonly database"},
-     { SQLITE_INTERRUPT, &Sqlite_Interrupt_Error, "Sqlite_Interrupt_Error", "Operation terminated by sqlite3_interrupt()*/"},
-     { SQLITE_IOERR, &Sqlite_Ioerr_Error, "Sqlite_Ioerr_Error", "Some kind of disk I/O error occurred"},
-     { SQLITE_CORRUPT, &Sqlite_Corrupt_Error, "Sqlite_Corrupt_Error", "The database disk image is malformed"},
-     { SQLITE_NOTFOUND, &Sqlite_Notfound_Error, "Sqlite_Notfound_Error", "NOT USED. Table or record not found"},
-     { SQLITE_FULL, &Sqlite_Full_Error, "Sqlite_Full_Error", "Insertion failed because database is full"},
-     { SQLITE_CANTOPEN, &Sqlite_Cantopen_Error, "Sqlite_Cantopen_Error", "Unable to open the database file"},
-     { SQLITE_PROTOCOL, &Sqlite_Protocol_Error, "Sqlite_Protocol_Error", "Database lock protocol error"},
-     { SQLITE_EMPTY, &Sqlite_Empty_Error, "Sqlite_Empty_Error", "Database is empty"},
-     { SQLITE_SCHEMA, &Sqlite_Schema_Error, "Sqlite_Schema_Error", "The database schema changed"},
-     { SQLITE_TOOBIG, &Sqlite_Toobig_Error, "Sqlite_Toobig_Error", "NOT USED. Too much data for one row"},
-     { SQLITE_CONSTRAINT, &Sqlite_Constraint_Error, "Sqlite_Constraint_Error", "Abort due to contraint violation"},
-     { SQLITE_MISMATCH, &Sqlite_Mismatch_Error, "Sqlite_Mismatch_Error", "Data type mismatch"},
-     { SQLITE_MISUSE, &Sqlite_Misuse_Error, "Sqlite_Misuse_Error", "Library used incorrectly"},
-     { SQLITE_NOLFS, &Sqlite_Nolfs_Error, "Sqlite_Nolfs_Error", "Uses OS features not supported on host"},
-     { SQLITE_AUTH, &Sqlite_Auth_Error, "Sqlite_Auth_Error", "Authorization denied"},
-     { SQLITE_FORMAT, &Sqlite_Format_Error, "Sqlite_Format_Error", "Auxiliary database format error"},
-     { SQLITE_RANGE, &Sqlite_Range_Error, "Sqlite_Range_Error", "2nd parameter to sqlite3_bind out of range"},
-     { SQLITE_NOTADB, &Sqlite_Notadb_Error, "Sqlite_Notadb_Error", "File opened that is not a database file"},
+     { SQLITE_ERROR, &Sqlite_Error, "SqliteError", "Sqlite Error"},
+     { SQLITE_INTERNAL, &Sqlite_Internal_Error, "SqliteInternalError", "NOT USED. Internal logic error in SQLite"},
+     { SQLITE_PERM, &Sqlite_Perm_Error, "SqlitePermError", "Access permission denied"},
+     { SQLITE_ABORT, &Sqlite_Abort_Error, "SqliteAbortError", "Callback routine requested an abort"},
+     { SQLITE_BUSY, &Sqlite_Busy_Error, "SqliteBusyError", "The database file is locked"},
+     { SQLITE_LOCKED, &Sqlite_Locked_Error, "SqliteLockedError", "A table in the database is locked"},
+     { SQLITE_NOMEM, &Sqlite_Nomem_Error, "SqliteNomemError", "A malloc() failed"},
+     { SQLITE_READONLY, &Sqlite_Readonly_Error, "SqliteReadonlyError", "Attempt to write a readonly database"},
+     { SQLITE_INTERRUPT, &Sqlite_Interrupt_Error, "SqliteInterruptError", "Operation terminated by sqlite3_interrupt()*/"},
+     { SQLITE_IOERR, &Sqlite_Ioerr_Error, "SqliteIoerrError", "Some kind of disk I/O error occurred"},
+     { SQLITE_CORRUPT, &Sqlite_Corrupt_Error, "SqliteCorruptError", "The database disk image is malformed"},
+     { SQLITE_NOTFOUND, &Sqlite_Notfound_Error, "SqliteNotfoundError", "NOT USED. Table or record not found"},
+     { SQLITE_FULL, &Sqlite_Full_Error, "SqliteFullError", "Insertion failed because database is full"},
+     { SQLITE_CANTOPEN, &Sqlite_Cantopen_Error, "SqliteCantopenError", "Unable to open the database file"},
+     { SQLITE_PROTOCOL, &Sqlite_Protocol_Error, "SqliteProtocolError", "Database lock protocol error"},
+     { SQLITE_EMPTY, &Sqlite_Empty_Error, "SqliteEmptyError", "Database is empty"},
+     { SQLITE_SCHEMA, &Sqlite_Schema_Error, "SqliteSchemaError", "The database schema changed"},
+     { SQLITE_TOOBIG, &Sqlite_Toobig_Error, "SqliteToobigError", "NOT USED. Too much data for one row"},
+     { SQLITE_CONSTRAINT, &Sqlite_Constraint_Error, "SqliteConstraintError", "Abort due to contraint violation"},
+     { SQLITE_MISMATCH, &Sqlite_Mismatch_Error, "SqliteMismatchError", "Data type mismatch"},
+     { SQLITE_MISUSE, &Sqlite_Misuse_Error, "SqliteMisuseError", "Library used incorrectly"},
+     { SQLITE_NOLFS, &Sqlite_Nolfs_Error, "SqliteNolfsError", "Uses OS features not supported on host"},
+     { SQLITE_AUTH, &Sqlite_Auth_Error, "SqliteAuthError", "Authorization denied"},
+     { SQLITE_FORMAT, &Sqlite_Format_Error, "SqliteFormatError", "Auxiliary database format error"},
+     { SQLITE_RANGE, &Sqlite_Range_Error, "SqliteRangeError", "2nd parameter to sqlite3_bind out of range"},
+     { SQLITE_NOTADB, &Sqlite_Notadb_Error, "SqliteNotadbError", "File opened that is not a database file"},
      { SQLITE_OK, 0, 0, 0 }
 };
 
@@ -103,7 +125,7 @@ static int check_error (sqlite3 *db, int error_code)
 {
    const Sqlite_Exception_Table_Type *b;
    int error;
-   if (error_code == SQLITE_OK || error_code == SQLITE_DONE) return 0;
+   if (error_code == SQLITE_OK || error_code == SQLITE_DONE || error_code == SQLITE_ROW) return 0;
    b = Sqlite_Exception_Table;
    
    while (b->errcode_ptr != NULL)
@@ -114,12 +136,13 @@ static int check_error (sqlite3 *db, int error_code)
      }
    if (b->errcode_ptr == NULL) error = Sqlite_Error;
    else error = *(b->errcode_ptr);
-   SLang_verror (error, (char *) sqlite3_errmsg(db));
+   SLang_verror (error, "%s", sqlite3_errmsg(db));
    return 1;
 }
 
 /*}}}*/
 /*{{{ helper functions */
+/*{{{ sqlite type */
 
 static void free_db_type (db_type *pt)
 {
@@ -154,78 +177,138 @@ static SLang_MMT_Type *allocate_db_type (sqlite3 *db)
    return mmt;
 }
 
+/*}}}*/
+/*{{{ statement type */
+
+static void free_Statement_Type (Statement_Type *pt)
+{
+   if (pt->ppStmt != NULL)
+     sqlite3_finalize(pt->ppStmt);
+   SLfree ((char *) pt);
+}
+
+static void destroy_statement (SLtype type, VOID_STAR f)
+{
+   (void) type;
+   free_Statement_Type ((Statement_Type *) f);
+}
+
+static SLang_MMT_Type *allocate_statement_type (sqlite3_stmt *ppStmt)
+{
+   Statement_Type *pt;
+   SLang_MMT_Type *mmt;
+   
+   pt = (Statement_Type *) SLmalloc (sizeof (Statement_Type));
+   if (pt == NULL)
+     return NULL;
+   memset ((char *) pt, 0, sizeof (Statement_Type));
+   
+   pt->ppStmt = ppStmt;
+   pt->state  = SQLITE_OK;
+   
+   if (NULL == (mmt = SLang_create_mmt (Statement_Type_Id, (VOID_STAR) pt)))
+     {
+	free_Statement_Type (pt);
+	return NULL;
+     }   
+   return mmt;
+}
+
+/*}}}*/
+
+static void sqlite_fetch(sqlite3_stmt *ppStmt)
+{
+   int i;
+   SLang_BString_Type *bstr;
+   for (i = 0; i < sqlite3_data_count (ppStmt); i++)
+     {
+	switch (sqlite3_column_type(ppStmt, i))
+	  {
+	   case SQLITE_INTEGER:
+	     SLang_push_integer(sqlite3_column_int(ppStmt, i));
+	     break;
+	   case SQLITE_FLOAT:
+	     SLang_push_double(sqlite3_column_double(ppStmt, i));
+	     break;
+	   case SQLITE_TEXT:
+	     SLang_push_string((char *)sqlite3_column_text(ppStmt, i));
+	     break;
+	   case SQLITE_BLOB:
+	     bstr = SLbstring_create((unsigned char *)sqlite3_column_blob(ppStmt, i),
+				     sqlite3_column_bytes(ppStmt, i));
+	     if (bstr != NULL)
+	       (void) SLang_push_bstring(bstr);
+	     SLbstring_free(bstr);
+	     break;
+	   case SQLITE_NULL:
+	     SLang_push_null();
+	  }
+     }
+}
+
 static int sqlite_step(sqlite3 *db, sqlite3_stmt *ppStmt)
 {
-   int i, res;
-   SLang_BString_Type *bstr;
+   int res;
    res = sqlite3_step(ppStmt);
    if (res == SQLITE_ROW)
      {
-	for (i = 0; i < sqlite3_data_count (ppStmt); i++)
-	  {
-	     switch (sqlite3_column_type(ppStmt, i))
-	       {
-		case SQLITE_INTEGER:
-		  SLang_push_integer(sqlite3_column_int(ppStmt, i));
-		  break;
-		case SQLITE_FLOAT:
-		  SLang_push_double(sqlite3_column_double(ppStmt, i));
-		  break;
-		case SQLITE_TEXT:
-		  SLang_push_string((char *)sqlite3_column_text(ppStmt, i));
-		  break;
-		case SQLITE_BLOB:
-		  bstr = SLbstring_create((unsigned char *)sqlite3_column_blob(ppStmt, i),
-					  sqlite3_column_bytes(ppStmt, i));
-		  if (bstr != NULL)
-		    (void) SLang_push_bstring(bstr);
-		  SLbstring_free(bstr);
-		  break;
-		case SQLITE_NULL:
-		  SLang_push_null();
-	       }
-	  }
+	sqlite_fetch(ppStmt);
 	return 1;
      }
    if (res != SQLITE_DONE)
-     check_error(db, sqlite3_reset(ppStmt));
+     check_error(db, res);
    return -1;
 }
 
-static int sqlite_bind(sqlite3 *db, sqlite3_stmt *ppStmt, int num)
+static void slstring_destructor (void *p)
+{
+   SLang_free_slstring (p);
+}
+
+
+/*
+ * Bind parameters with positions i .. num.
+ * Usually, either i = 1 or i = num
+ */
+static int sqlite_bind(sqlite3 *db, sqlite3_stmt *ppStmt, int num, int i)
 {
-   int i;
-   int ivalue;
-   float fvalue;
-   double dvalue;
    char *svalue;
    SLang_BString_Type *bvalue;
    unsigned int bstrlen;
    unsigned char *bptr;
-   
-   for (i = 1; i <= num; i++)
+
+#define MAP(slangtype, ctype, slangpop, sqlitebind) \
+ case slangtype:\
+   {\
+      ctype value;\
+      slangpop(&value);\
+      if(check_error(db, sqlitebind(ppStmt, i, value)))\
+	return -1;\
+      break;\
+   }
+
+   for (; i <= num; i++)
      {
 	switch (SLang_peek_at_stack())
 	  {
-	   case SLANG_INT_TYPE:
-	     SLang_pop_int(&ivalue);
-	     if (check_error(db, sqlite3_bind_int(ppStmt, i, ivalue)))
-	       return -1;
-	     break;
-	   case SLANG_FLOAT_TYPE:
-	     SLang_pop_float(&fvalue);
-	     if (check_error(db, sqlite3_bind_double(ppStmt, i, (double) fvalue)))
-	       return -1;
-	     break;
-	   case SLANG_DOUBLE_TYPE:
-	     SLang_pop_double(&dvalue);
-	     if (check_error(db, sqlite3_bind_double(ppStmt, i, dvalue)))
-	       return -1;
-	     break;
+	     MAP(SLANG_INT_TYPE, int, SLang_pop_int, sqlite3_bind_int);
+	     MAP(SLANG_FLOAT_TYPE, float, SLang_pop_float, sqlite3_bind_double);
+	     MAP(SLANG_DOUBLE_TYPE, double, SLang_pop_double, sqlite3_bind_double);
+#ifdef HAVE_LONG_LONG
+	   case SLANG_UINT_TYPE:
+	       {
+		  unsigned int value;
+		  SLang_pop_uint(&value);
+		  if (check_error(db, sqlite3_bind_int64(ppStmt, i, (long long int) value)))
+		    return -1;
+		  break;
+	       }
+	     MAP(SLANG_LLONG_TYPE, long long, SLang_pop_long_long, sqlite3_bind_int64);
+#endif
 	   case SLANG_STRING_TYPE:
 	     SLang_pop_slstring(&svalue);
 	     if (check_error(db, sqlite3_bind_text(ppStmt, i, svalue,
-						   strlen(svalue), (void *)&SLang_free_slstring)))
+                                                  strlen(svalue), slstring_destructor)))
 	       return -1;
 	     break;
 	   case SLANG_BSTRING_TYPE:
@@ -245,6 +328,7 @@ static int sqlite_bind(sqlite3 *db, sqlite3_stmt *ppStmt, int num)
 	     return -1;
 	  }
      }
+#undef MAP
    return 0;
 }
 
@@ -272,6 +356,38 @@ static void slsqlite_open(char *name)
      }
 }
 
+static void slsqlite_prepare (const char *sql)
+{
+   db_type *p;
+   SLang_MMT_Type *mmt, *mmt2;
+   sqlite3_stmt *ppStmt;
+   if (NULL == (mmt = SLang_pop_mmt (DB_Type_Id)))
+     {
+	SLang_free_mmt (mmt);
+	(void) SLang_push_null();
+	return;
+     }
+   p = SLang_object_from_mmt (mmt);
+   
+   if (check_error(p->db, sqlite3_prepare_v2(p->db, sql, -1, &ppStmt, NULL)))
+     {
+	goto free_return;
+     }
+   if (NULL == (mmt2 = allocate_statement_type (ppStmt)))
+     {
+	(void) SLang_push_null();
+	goto free_return;
+     }
+   if (-1 == SLang_push_mmt (mmt2))
+     {
+	SLang_free_mmt (mmt2);
+	goto free_return;
+     }
+   
+free_return:
+   SLang_free_mmt (mmt);
+}
+
 static void slsqlite_get_table (const char *sql)
 {
    db_type *p;
@@ -301,7 +417,7 @@ static void slsqlite_get_table (const char *sql)
    
    if (at != NULL)
      {
-	char **p = resultp;
+       char **pp = resultp;
 	SLindex_Type j, k, d[2];
 	for (j=0; j<dims[0]; j++)
 	  {
@@ -309,7 +425,7 @@ static void slsqlite_get_table (const char *sql)
 	     for (k=0; k<dims[1]; k++)
 	       {
 		  d[1]=k;
-		  SLang_set_array_element (at, d, p++);
+                 SLang_set_array_element (at, d, pp++);
 	       }
 	  }
 	
@@ -336,6 +452,7 @@ static void slsqlite_get_row(void)
    
    if (nargs < 2)
      {
+       (void) SLdo_pop_n (nargs);
 	SLang_verror(SL_Usage_Error, "usage: sqlite_get_row(Sqlite db, String sql, ...)");
 	return;
      }
@@ -355,12 +472,12 @@ static void slsqlite_get_row(void)
 	return;
      }
    
-   if (check_error(p->db, sqlite3_prepare(p->db, sql, strlen(sql), &ppStmt, NULL)))
+   if (check_error(p->db, sqlite3_prepare_v2(p->db, sql, -1, &ppStmt, NULL)))
      {
 	goto free_return;
      }
 
-   if (sqlite_bind(p->db, ppStmt, nargs - 2))
+   if (sqlite_bind(p->db, ppStmt, nargs - 2, 1))
      {
 	sqlite3_finalize(ppStmt);
 	goto free_return;
@@ -379,128 +496,72 @@ free_return:
 
 /*{{{ sqlite_get_array */
 
-static void sqlite_get_integer_array(sqlite3_stmt *ppStmt)
-{
-   unsigned int num_items = 0, max_num_items = 1024;
-   SLindex_Type dims[2];
-   int *list;
-   SLang_Array_Type *at;
-   list = (int *) SLmalloc (sizeof (int) * max_num_items);
-   if (list == NULL)
-     {
-	SLang_verror(SL_Malloc_Error, "Out of memory");
-	return;
-     }
-   dims[0] = 0;
-
-   while (SQLITE_ROW == sqlite3_step(ppStmt))
-     {
-	int i;
-	dims[0]++;
-	dims[1] = sqlite3_data_count(ppStmt);
-
-	for (i = 0; i < sqlite3_data_count (ppStmt); i++)
-	  {
-	     int value = sqlite3_column_int(ppStmt, i);
-	     if (max_num_items == num_items)
-	       {
-		  int *new_list;
-		  max_num_items += 4096;
-		  
-		  new_list = (int *) SLrealloc ((char *)list, sizeof (int) * max_num_items);
-		  if (new_list == NULL)
-		    {
-		       SLang_verror(SL_Malloc_Error, "Out of memory");
-		       goto return_error;
-		    }
-		  list = new_list;
-	       }
-	     
-	     list[num_items] = value;
-	     num_items++;
-	  }
-     }
-   if (num_items != max_num_items)
-     {
-	int *new_list = (int *)SLrealloc ((char *)list, sizeof (int) * (num_items + 1));
-	if (new_list == NULL)
-	  {
-	     SLang_verror(SL_Malloc_Error, "Out of memory");
-	     goto return_error;
-	  }
-	list = new_list;
-     }
-   if ((NULL == (at = SLang_create_array (SLANG_INT_TYPE, 0, (VOID_STAR) list, dims, 2)))
-       || (-1 == SLang_push_array (at, 1)))
-     SLang_push_null ();
-   return;
-   
-return_error:
-   SLfree ((char *)list);
+#define GET_ARRAY(type, ctype, sqltype, slangtype)\
+static void sqlite_get_##type##_array(sqlite3_stmt *ppStmt)\
+{\
+   unsigned int num_items = 0, max_num_items = 1024;\
+   SLindex_Type dims[2];\
+   ctype *list;\
+   SLang_Array_Type *at;\
+   list = (ctype *) SLmalloc (sizeof (ctype) * max_num_items);\
+   if (list == NULL)\
+     {\
+	SLang_verror(SL_Malloc_Error, "Out of memory");\
+	return;\
+     }\
+   dims[1] = dims[0] = 0;\
+\
+   while (SQLITE_ROW == sqlite3_step(ppStmt))\
+     {\
+	ctype i;\
+	dims[0]++;\
+	dims[1] = sqlite3_data_count(ppStmt);\
+\
+	for (i = 0; i < sqlite3_data_count (ppStmt); i++)\
+	  {\
+	     ctype value = sqlite3_column_##sqltype (ppStmt, i);\
+	     if (max_num_items == num_items)\
+	       {\
+		  ctype *new_list;\
+		  max_num_items += 4096;\
+		  \
+		  new_list = (ctype *) SLrealloc ((char *)list, sizeof (ctype) * max_num_items);\
+		  if (new_list == NULL)\
+		    {\
+		       SLang_verror(SL_Malloc_Error, "Out of memory");\
+		       goto return_error;\
+		    }\
+		  list = new_list;\
+	       }\
+	     \
+	     list[num_items] = value;\
+	     num_items++;\
+	  }\
+     }\
+   if (num_items != max_num_items)\
+     {\
+	ctype *new_list = (ctype *)SLrealloc ((char *)list, sizeof (ctype) * (num_items + 1));\
+	if (new_list == NULL)\
+	  {\
+	     SLang_verror(SL_Malloc_Error, "Out of memory");\
+	     goto return_error;\
+	  }\
+	list = new_list;\
+     }\
+   if ((NULL == (at = SLang_create_array (slangtype, 0, (VOID_STAR) list, dims, 2)))\
+       || (-1 == SLang_push_array (at, 1)))\
+     SLang_push_null ();\
+   return;\
+   \
+return_error:\
+   SLfree ((char *)list);\
 }
 
-static void sqlite_get_double_array(sqlite3_stmt *ppStmt)
-{
-   unsigned int num_items = 0, max_num_items = 1024;
-   SLindex_Type dims[2];
-   double *list;
-   SLang_Array_Type *at;
-   list = (double *) SLmalloc (sizeof (double) * max_num_items);
-   if (list == NULL)
-     {
-	SLang_verror(SL_Malloc_Error, "Out of memory");
-	return;
-     }
-
-   dims[0] = 0;
-   
-   while (SQLITE_ROW == sqlite3_step(ppStmt))
-     {
-	int i;
-	dims[0]++;
-	dims[1] = sqlite3_data_count(ppStmt);
-	
-	for (i = 0; i < sqlite3_data_count (ppStmt); i++)
-	  {
-	     double value = sqlite3_column_double(ppStmt, i);
-	     if (max_num_items == num_items)
-	       {
-		  double *new_list;
-		  max_num_items += 4096;
-		  
-		  new_list = (double *) SLrealloc ((char *)list, sizeof (double) * max_num_items);
-		  if (new_list == NULL)
-		    {
-		       SLang_verror(SL_Malloc_Error, "Out of memory");
-		       goto return_error;
-		    }
-		  list = new_list;
-	       }
-	     
-	     list[num_items] = value;
-	     num_items++;
-	  }
-	
-     }
-   
-   if (num_items != max_num_items)
-     {
-	double *new_list = (double *)SLrealloc ((char *)list, sizeof (double) * (num_items + 1));
-	if (new_list == NULL)
-	  {
-	     SLang_verror(SL_Malloc_Error, "Out of memory");
-	     goto return_error;
-	  }
-	list = new_list;
-     }
-   if ((NULL == (at = SLang_create_array (SLANG_DOUBLE_TYPE, 0, (VOID_STAR) list, dims, 2)))
-       || (-1 == SLang_push_array (at, 1)))
-     SLang_push_null ();
-   return;
-   
-return_error:
-   SLfree ((char *)list);
-}
+GET_ARRAY(integer, int, int, SLANG_INT_TYPE);
+GET_ARRAY(double, double, double, SLANG_DOUBLE_TYPE);
+#ifdef HAVE_LONG_LONG
+GET_ARRAY(llong, long long, int64, SLANG_LLONG_TYPE);
+#endif
 
 static void sqlite_get_string_array(sqlite3_stmt *ppStmt)
 {
@@ -515,7 +576,7 @@ static void sqlite_get_string_array(sqlite3_stmt *ppStmt)
 	return;
      }
    
-   dims[0] = 0;
+   dims[1] = dims[0] = 0;
 
    while (SQLITE_ROW == sqlite3_step(ppStmt))
      {
@@ -570,6 +631,7 @@ return_error:
      }
    SLfree ((char *)list);
 }
+
 static void sqlite_get_bstring_array(sqlite3_stmt *ppStmt)
 {
    unsigned int num_items = 0, max_num_items = 1024;
@@ -585,7 +647,7 @@ static void sqlite_get_bstring_array(sqlite3_stmt *ppStmt)
 	return;
      }
    
-   dims[0] = 0;
+   dims[1] = dims[0] = 0;
 
    while (SQLITE_ROW == sqlite3_step(ppStmt))
      {
@@ -654,8 +716,8 @@ static void slsqlite_get_array(void)
    
    if (nargs < 3)
      {
+       SLdo_pop_n (nargs);
 	SLang_verror(SL_Usage_Error, "usage: sqlite_get_array(Sqlite db, DataType type, String sql, ...)");
-	SLdo_pop_n (SLang_Num_Function_Args);
 	return;
      }
    
@@ -682,13 +744,16 @@ static void slsqlite_get_array(void)
    switch(type)
      {
       case SLANG_INT_TYPE:
+#ifdef HAVE_LONG_LONG
+      case SLANG_LLONG_TYPE:
+#endif
       case SLANG_DOUBLE_TYPE:
       case SLANG_STRING_TYPE:
       case SLANG_BSTRING_TYPE:
 	break;
       default:
-	SLang_verror(SL_Usage_Error, "only Integer, Double, String and Bstring types allowed");
 	SLdo_pop_n (nargs - 1);
+	SLang_verror(SL_Usage_Error, "only Integer, Double, String and Bstring types allowed");
 	return;
      }
    
@@ -700,12 +765,12 @@ static void slsqlite_get_array(void)
 	return;
      }
    
-   if (check_error(p->db, sqlite3_prepare(p->db, sql, strlen(sql), &ppStmt, NULL)))
+   if (check_error(p->db, sqlite3_prepare_v2(p->db, sql, -1, &ppStmt, NULL)))
      {
 	goto free_return;
      }
    
-   if (sqlite_bind(p->db, ppStmt, nargs - 2))
+   if (sqlite_bind(p->db, ppStmt, nargs - 2, 1))
      {
 	sqlite3_finalize(ppStmt);
 	goto free_return;
@@ -717,6 +782,10 @@ static void slsqlite_get_array(void)
 	sqlite_get_integer_array(ppStmt); break;
       case SLANG_DOUBLE_TYPE:
 	sqlite_get_double_array(ppStmt); break;
+#ifdef HAVE_LONG_LONG
+      case SLANG_LLONG_TYPE:
+	sqlite_get_llong_array(ppStmt); break;
+#endif
       case SLANG_STRING_TYPE:
 	sqlite_get_string_array(ppStmt); break;
       case SLANG_BSTRING_TYPE:
@@ -743,16 +812,15 @@ static void slsqlite_exec(void)
    
    if (nargs < 2)
      {
+       (void) SLdo_pop_n (nargs);
 	SLang_verror(SL_Usage_Error, "usage: sqlite_exec(Sqlite db, String sql, ...)");
 	return;
      }
    
    SLreverse_stack(nargs);
    if (NULL == (mmt = SLang_pop_mmt (DB_Type_Id)))
-     {
-	SLang_free_mmt (mmt);
-	return;
-     }  
+     return;
+
    p = SLang_object_from_mmt (mmt);
    
    if (-1 == SLang_pop_slstring(&sql))
@@ -762,22 +830,19 @@ static void slsqlite_exec(void)
 	return;
      }
    
-   if (check_error(p->db, sqlite3_prepare(p->db, sql, strlen(sql), &ppStmt, NULL)))
+   if (check_error(p->db, sqlite3_prepare_v2(p->db, sql, -1, &ppStmt, NULL)))
      {
 	goto free_return;
      }
    
-   if (sqlite_bind(p->db, ppStmt, nargs - 2))
+   if (sqlite_bind(p->db, ppStmt, nargs - 2, 1))
      {
 	sqlite3_finalize(ppStmt);
 	goto free_return;
      }
    
-   if (SQLITE_ERROR == sqlite3_step(ppStmt))
-     {
-	check_error(p->db, sqlite3_reset(ppStmt));
-	sqlite3_finalize(ppStmt);
-     }
+   if (check_error(p->db, sqlite3_step(ppStmt)))
+     sqlite3_finalize(ppStmt);
    else
      check_error(p->db, sqlite3_finalize (ppStmt));
    
@@ -786,7 +851,7 @@ free_return:
    SLang_free_mmt(mmt);
 }
 
-static int slsqlite_changes()
+static int slsqlite_changes (void)
 {
    db_type *p;
    SLang_MMT_Type *mmt;
@@ -802,17 +867,196 @@ static int slsqlite_changes()
    return res;
 }
 
+static void slsqlite_bind_params (void)
+{
+   Statement_Type *p;
+   SLang_MMT_Type *mmt;
+   sqlite3 *db;
+   int nargs;
+   
+   nargs = SLang_Num_Function_Args;
+   
+   if (nargs < 1)
+     {
+	SLdo_pop_n (nargs);
+	SLang_verror(SL_Usage_Error, "usage: sqlite_bind_params(Statement stmt, ...)");
+	return;
+     }
+   
+   SLreverse_stack(nargs);
+   
+   if (NULL == (mmt = SLang_pop_mmt (Statement_Type_Id)))
+     {
+	SLang_free_mmt (mmt);
+	return;
+     }
+   
+   nargs--;
+   
+   p = SLang_object_from_mmt (mmt);
+   
+   /* 
+    * This is necessary because the documentation for sqlite_bind_* says
+    * 
+    * SQLITE_MISUSE might be returned if these routines are called on a
+    * virtual machine that is the wrong state or which has already been
+    * finalized. [ ... ] Future versions of SQLite might panic rather than
+    * return SQLITE_MISUSE.
+    */
+   if (p->state != SQLITE_OK)
+     {
+	SLang_verror (Sqlite_Error, "prepared statement is in wrong state (%d)", p->state);
+	goto free_return;
+     }
+
+   db = sqlite3_db_handle(p->ppStmt);
+   sqlite_bind(db, p->ppStmt, nargs, 1);
+free_return:
+   SLang_free_mmt(mmt);
+}
+
+/*
+ * The next two function are needed to add support for named parameters in
+ * the 'bind_params' method
+ */
+
+static void slsqlite_bind_param (void)
+{
+   Statement_Type *p;
+   SLang_MMT_Type *mmt;
+   sqlite3 *db;
+   int nargs;
+   int n;
+   
+   nargs = SLang_Num_Function_Args;
+   
+   if (nargs < 1)
+     {
+	SLdo_pop_n (nargs);
+	SLang_verror(SL_Usage_Error, "usage: sqlite_bind_param(Statement stmt, int n, value)");
+	return;
+     }
+   
+   SLreverse_stack(nargs);
+   
+   if (NULL == (mmt = SLang_pop_mmt (Statement_Type_Id))
+       || (SLang_pop_int (&n)))
+     {
+	SLang_free_mmt (mmt);
+	return;
+     }
+   
+   p = SLang_object_from_mmt (mmt);
+   if (p->state != SQLITE_OK)
+     {
+	SLang_verror (Sqlite_Error, "prepared statement is in wrong state (%d)", p->state);
+	goto free_return;
+     }
+
+   db = sqlite3_db_handle(p->ppStmt);
+   sqlite_bind(db, p->ppStmt, n, n);
+free_return:
+   SLang_free_mmt(mmt);
+}
+
+static int slsqlite_bind_parameter_index (const char *name)
+{
+   Statement_Type *p;
+   SLang_MMT_Type *mmt;
+   int res = 0;
+   
+   if (NULL == (mmt = SLang_pop_mmt (Statement_Type_Id)))
+     {
+	SLang_free_mmt (mmt);
+	return 0;
+     }
+   
+   p = SLang_object_from_mmt (mmt);
+   res = sqlite3_bind_parameter_index (p->ppStmt, name);
+   SLang_free_mmt(mmt);
+   return res;
+}
+
+static int slsqlite_step (void)
+{
+   Statement_Type *p;
+   SLang_MMT_Type *mmt;
+   int res = 0;
+   
+   if (NULL == (mmt = SLang_pop_mmt (Statement_Type_Id)))
+     {
+	SLang_free_mmt (mmt);
+	return 0;
+     }
+   
+   p = SLang_object_from_mmt (mmt);
+   if (p->state != SQLITE_OK && p->state != SQLITE_ROW)
+     {
+	/* 
+	 * This can occur when trying to step in a ppStmt that has finished
+	 */
+	SLang_verror (Sqlite_Error, "prepared statement is in wrong state (%d)", p->state);
+     }
+   res = sqlite3_step (p->ppStmt);
+   (void) check_error (sqlite3_db_handle (p->ppStmt), res);
+   p->state = res;
+   SLang_free_mmt (mmt);
+   return res;
+}
+
+static void slsqlite_fetch(void)
+{
+   Statement_Type *p;
+   SLang_MMT_Type *mmt;
+   
+   if (NULL == (mmt = SLang_pop_mmt (Statement_Type_Id)))
+     {
+	SLang_free_mmt (mmt);
+	return;
+     }
+   p = SLang_object_from_mmt (mmt);
+   p = SLang_object_from_mmt (mmt);
+   if (p->state != SQLITE_ROW)
+     {
+	SLang_verror (Sqlite_Error, "prepared statement is in wrong state (%d)", p->state);
+     }
+   sqlite_fetch(p->ppStmt);
+   SLang_free_mmt(mmt);
+}
+
+static void slsqlite_reset(void)
+{
+   Statement_Type *p;
+   SLang_MMT_Type *mmt;
+   
+   if (NULL == (mmt = SLang_pop_mmt (Statement_Type_Id)))
+     {
+	SLang_free_mmt (mmt);
+	return;
+     }
+   p = SLang_object_from_mmt (mmt);
+   p->state = sqlite3_reset(p->ppStmt);
+   SLang_free_mmt(mmt);
+}
+
 /*}}}*/
 /*{{{ intrinsics */
 
 static SLang_Intrin_Fun_Type Module_Intrinsics [] =
 {
    MAKE_INTRINSIC_S("sqlite_open", slsqlite_open, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("sqlite_prepare", slsqlite_prepare, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("sqlite_get_table", slsqlite_get_table, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("sqlite_get_row", slsqlite_get_row, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("sqlite_get_array", slsqlite_get_array, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("sqlite_exec", slsqlite_exec, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("sqlite_changes", slsqlite_changes, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("sqlite_bind_params", slsqlite_bind_params, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("sqlite_bind_param", slsqlite_bind_param, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("sqlite_bind_parameter_index", slsqlite_bind_parameter_index, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("sqlite_step", slsqlite_step, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("sqlite_fetch", slsqlite_fetch, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("sqlite_reset", slsqlite_reset, SLANG_VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
@@ -825,6 +1069,35 @@ static SLang_Intrin_Var_Type Module_Variables [] =
 static SLang_IConstant_Type Module_Constants [] =
 {
    MAKE_ICONSTANT("_sqlite_module_version", MODULE_VERSION_NUMBER),
+   MAKE_ICONSTANT("SQLITE_OK", SQLITE_OK),
+   MAKE_ICONSTANT("SQLITE_ERROR", SQLITE_ERROR),
+   MAKE_ICONSTANT("SQLITE_INTERNAL", SQLITE_INTERNAL),
+   MAKE_ICONSTANT("SQLITE_PERM", SQLITE_PERM),
+   MAKE_ICONSTANT("SQLITE_ABORT", SQLITE_ABORT),
+   MAKE_ICONSTANT("SQLITE_BUSY", SQLITE_BUSY),
+   MAKE_ICONSTANT("SQLITE_LOCKED", SQLITE_LOCKED),
+   MAKE_ICONSTANT("SQLITE_NOMEM", SQLITE_NOMEM),
+   MAKE_ICONSTANT("SQLITE_READONLY", SQLITE_READONLY),
+   MAKE_ICONSTANT("SQLITE_INTERRUPT", SQLITE_INTERRUPT),
+   MAKE_ICONSTANT("SQLITE_IOERR", SQLITE_IOERR),
+   MAKE_ICONSTANT("SQLITE_CORRUPT", SQLITE_CORRUPT),
+   MAKE_ICONSTANT("SQLITE_NOTFOUND", SQLITE_NOTFOUND),
+   MAKE_ICONSTANT("SQLITE_FULL", SQLITE_FULL),
+   MAKE_ICONSTANT("SQLITE_CANTOPEN", SQLITE_CANTOPEN),
+   MAKE_ICONSTANT("SQLITE_PROTOCOL", SQLITE_PROTOCOL),
+   MAKE_ICONSTANT("SQLITE_EMPTY", SQLITE_EMPTY),
+   MAKE_ICONSTANT("SQLITE_SCHEMA", SQLITE_SCHEMA),
+   MAKE_ICONSTANT("SQLITE_TOOBIG", SQLITE_TOOBIG),
+   MAKE_ICONSTANT("SQLITE_CONSTRAINT", SQLITE_CONSTRAINT),
+   MAKE_ICONSTANT("SQLITE_MISMATCH", SQLITE_MISMATCH),
+   MAKE_ICONSTANT("SQLITE_MISUSE", SQLITE_MISUSE),
+   MAKE_ICONSTANT("SQLITE_NOLFS", SQLITE_NOLFS),
+   MAKE_ICONSTANT("SQLITE_AUTH", SQLITE_AUTH),
+   MAKE_ICONSTANT("SQLITE_FORMAT", SQLITE_FORMAT),
+   MAKE_ICONSTANT("SQLITE_RANGE", SQLITE_RANGE),
+   MAKE_ICONSTANT("SQLITE_NOTADB", SQLITE_NOTADB),
+   MAKE_ICONSTANT("SQLITE_ROW", SQLITE_ROW),
+   MAKE_ICONSTANT("SQLITE_DONE", SQLITE_DONE),
    SLANG_END_ICONST_TABLE
 };
 
@@ -872,10 +1145,10 @@ static SLang_Foreach_Context_Type *cl_foreach_open (SLtype type, unsigned int nu
    
    c->mmt = mmt;
    c->p = (db_type *) SLang_object_from_mmt (mmt);
-   if (check_error(c->p->db, sqlite3_prepare(c->p->db, s, strlen(s), &c->ppStmt, NULL)))
+   if (check_error(c->p->db, sqlite3_prepare_v2(c->p->db, s, -1, &c->ppStmt, NULL)))
      goto free_return;
    
-   if (sqlite_bind(c->p->db, c->ppStmt, num - 1))
+   if (sqlite_bind(c->p->db, c->ppStmt, num - 1, 1))
      {
 	sqlite3_finalize(c->ppStmt);
 	goto free_return;
@@ -964,6 +1237,32 @@ static int register_sqlite_type (void)
    return 0;
 }
 
+static int register_statement_type (void)
+{
+   SLang_Class_Type *cl;
+   
+   if (Statement_Type_Id != 0)
+     return 0;
+
+   if (NULL == (cl = SLclass_allocate_class ("Sqlite_Statement_Type")))
+     return -1;
+
+   if (-1 == SLclass_set_destroy_function (cl, destroy_statement))
+     return -1;
+
+   /* By registering as SLANG_VOID_TYPE, slang will dynamically allocate a
+    * type.
+    */
+   if (-1 == SLclass_register_class (cl, SLANG_VOID_TYPE, sizeof (Statement_Type), SLANG_CLASS_TYPE_MMT))
+     return -1;
+
+   Statement_Type_Id = SLclass_get_class_id (cl);
+
+   patchup_intrinsic_table (Module_Intrinsics, DUMMY_SQLITE_TYPE, Statement_Type_Id);
+   
+   return 0;
+}
+
 /*}}}*/
 /*{{{ init */
 
@@ -972,7 +1271,8 @@ int init_sqlite_module_ns (char *ns_name)
    SLang_NameSpace_Type *ns = SLns_create_namespace (ns_name);
    if (ns == NULL)
      return -1;
-   if (-1 == register_sqlite_type ())
+   if ((-1 == register_sqlite_type ())
+       || (-1 == register_statement_type ())) 
      return -1;
    
    if (Sqlite_Error == 0)
